@@ -22,36 +22,31 @@ from warnings import warn
 from weakref import WeakKeyDictionary, WeakValueDictionary
 
 try:
-    from typing import Literal, TypedDict
+    from typing_extensions import Literal, NoReturn
 except ImportError:
-    Literal = TypedDict = None
-
-try:
-    from typing_extensions import Literal as BPLiteral
-except ImportError:
-    BPLiteral = None
+    from typing import Literal, NoReturn
 
+# Python 3.6+
 try:
+    from inspect import isasyncgenfunction, isasyncgen
     from typing import AsyncGenerator
 except ImportError:
     AsyncGenerator = None
 
-try:
-    from typing import ForwardRef
-    evaluate_forwardref = ForwardRef._evaluate
-except ImportError:
-    from typing import _ForwardRef as ForwardRef  # Python < 3.8
-    evaluate_forwardref = ForwardRef._eval_type
-
-try:
-    from inspect import isasyncgenfunction, isasyncgen
-except ImportError:
     def isasyncgen(obj):
         return False
 
     def isasyncgenfunction(func):
         return False
 
+# Python 3.8+
+try:
+    from typing import ForwardRef
+    evaluate_forwardref = ForwardRef._evaluate
+except ImportError:
+    from typing import _ForwardRef as ForwardRef
+    evaluate_forwardref = ForwardRef._eval_type
+
 
 _type_hints_map = WeakKeyDictionary()  # type: Dict[FunctionType, Dict[str, Any]]
 _functions_map = WeakValueDictionary()  # type: Dict[CodeType, FunctionType]
@@ -77,21 +72,29 @@ class TypeHintWarning(UserWarning):
     """
 
 
-class _CallMemo:
-    __slots__ = ('func', 'func_name', 'signature', 'typevars', 'arguments', 'type_hints',
-                 'is_generator')
+class _TypeCheckMemo:
+    __slots__ = 'globals', 'locals', 'typevars'
+
+    def __init__(self, globals: Dict[str, Any], locals: Dict[str, Any]):
+        self.globals = globals
+        self.locals = locals
+        self.typevars = {}  # type: Dict[Any, type]
+
+
+class _CallMemo(_TypeCheckMemo):
+    __slots__ = 'func', 'func_name', 'arguments', 'is_generator', 'type_hints'
 
     def __init__(self, func: Callable, frame_locals: Optional[Dict[str, Any]] = None,
                  args: tuple = None, kwargs: Dict[str, Any] = None,
                  forward_refs_policy=ForwardRefPolicy.ERROR):
+        super().__init__(func.__globals__, frame_locals)
         self.func = func
         self.func_name = function_name(func)
-        self.signature = inspect.signature(func)
-        self.typevars = {}  # type: Dict[Any, type]
         self.is_generator = isgeneratorfunction(func)
+        signature = inspect.signature(func)
 
         if args is not None and kwargs is not None:
-            self.arguments = self.signature.bind(*args, **kwargs).arguments
+            self.arguments = signature.bind(*args, **kwargs).arguments
         else:
             assert frame_locals is not None, 'frame must be specified if args or kwargs is None'
             self.arguments = frame_locals
@@ -109,7 +112,7 @@ class _CallMemo:
                         raise
 
                     typename = str(exc).split("'", 2)[1]
-                    for param in self.signature.parameters.values():
+                    for param in signature.parameters.values():
                         if param.annotation == typename:
                             break
                     else:
@@ -134,7 +137,7 @@ class _CallMemo:
                     break
 
             self.type_hints = OrderedDict()
-            for name, parameter in self.signature.parameters.items():
+            for name, parameter in signature.parameters.items():
                 if name in hints:
                     annotated_type = hints[name]
 
@@ -155,16 +158,16 @@ class _CallMemo:
             _type_hints_map[func] = self.type_hints
 
 
-def resolve_forwardref(maybe_ref, memo: _CallMemo):
+def resolve_forwardref(maybe_ref, memo: _TypeCheckMemo):
     if isinstance(maybe_ref, ForwardRef):
-        return evaluate_forwardref(maybe_ref, memo.func.__globals__, {})
+        return evaluate_forwardref(maybe_ref, memo.globals, memo.locals)
     else:
         return maybe_ref
 
 
 def get_type_name(type_):
     # typing.* types don't have a __name__ on Python 3.7+
-    return getattr(type_, '__name__', None) or type_._name or str(type_)
+    return getattr(type_, '__name__', None) or getattr(type_, '_name', None) or str(type_)
 
 
 def find_function(frame) -> Optional[Callable]:
@@ -227,7 +230,7 @@ def function_name(func: Callable) -> str:
     return qualname if module == 'builtins' else '{}.{}'.format(module, qualname)
 
 
-def check_callable(argname: str, value, expected_type, memo: Optional[_CallMemo]) -> None:
+def check_callable(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
     if not callable(value):
         raise TypeError('{} must be a callable'.format(argname))
 
@@ -275,7 +278,7 @@ def check_callable(argname: str, value, expected_type, memo: Optional[_CallMemo]
                                                          num_mandatory_args))
 
 
-def check_dict(argname: str, value, expected_type, memo: Optional[_CallMemo]) -> None:
+def check_dict(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
     if not isinstance(value, dict):
         raise TypeError('type of {} must be a dict; got {} instead'.
                         format(argname, qualified_name(value)))
@@ -289,7 +292,7 @@ def check_dict(argname: str, value, expected_type, memo: Optional[_CallMemo]) ->
                     check_type('{}[{!r}]'.format(argname, k), v, value_type, memo)
 
 
-def check_typed_dict(argname: str, value, expected_type, memo: Optional[_CallMemo]) -> None:
+def check_typed_dict(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
     expected_keys = frozenset(expected_type.__annotations__)
     existing_keys = frozenset(value)
 
@@ -304,13 +307,13 @@ def check_typed_dict(argname: str, value, expected_type, memo: Optional[_CallMem
             keys_formatted = ', '.join('"{}"'.format(key) for key in sorted(missing_keys))
             raise TypeError('required key(s) ({}) missing from {}'.format(keys_formatted, argname))
 
-    for key, argtype in expected_type.__annotations__.items():
+    for key, argtype in get_type_hints(expected_type).items():
         argvalue = value.get(key, _missing)
         if argvalue is not _missing:
-            check_type('dict item "{}" for {}'.format(key, argname), argvalue, argtype)
+            check_type('dict item "{}" for {}'.format(key, argname), argvalue, argtype, memo)
 
 
-def check_list(argname: str, value, expected_type, memo: Optional[_CallMemo]) -> None:
+def check_list(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
     if not isinstance(value, list):
         raise TypeError('type of {} must be a list; got {} instead'.
                         format(argname, qualified_name(value)))
@@ -323,7 +326,7 @@ def check_list(argname: str, value, expected_type, memo: Optional[_CallMemo]) ->
                     check_type('{}[{}]'.format(argname, i), v, value_type, memo)
 
 
-def check_sequence(argname: str, value, expected_type, memo: Optional[_CallMemo]) -> None:
+def check_sequence(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
     if not isinstance(value, collections.abc.Sequence):
         raise TypeError('type of {} must be a sequence; got {} instead'.
                         format(argname, qualified_name(value)))
@@ -335,7 +338,7 @@ def check_sequence(argname: str, value, expected_type, memo: Optional[_CallMemo]
                 check_type('{}[{}]'.format(argname, i), v, value_type, memo)
 
 
-def check_set(argname: str, value, expected_type, memo: Optional[_CallMemo]) -> None:
+def check_set(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
     if not isinstance(value, AbstractSet):
         raise TypeError('type of {} must be a set; got {} instead'.
                         format(argname, qualified_name(value)))
@@ -348,7 +351,7 @@ def check_set(argname: str, value, expected_type, memo: Optional[_CallMemo]) ->
                     check_type('elements of {}'.format(argname), v, value_type, memo)
 
 
-def check_tuple(argname: str, value, expected_type, memo: Optional[_CallMemo]) -> None:
+def check_tuple(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
     # Specialized check for NamedTuples
     if hasattr(expected_type, '_field_types'):
         if not isinstance(value, expected_type):
@@ -391,7 +394,7 @@ def check_tuple(argname: str, value, expected_type, memo: Optional[_CallMemo]) -
             check_type('{}[{}]'.format(argname, i), element, element_type, memo)
 
 
-def check_union(argname: str, value, expected_type, memo: Optional[_CallMemo]) -> None:
+def check_union(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
     if hasattr(expected_type, '__union_params__'):
         # Python 3.5
         union_params = expected_type.__union_params__
@@ -411,7 +414,7 @@ def check_union(argname: str, value, expected_type, memo: Optional[_CallMemo]) -
                     format(argname, typelist, qualified_name(value)))
 
 
-def check_class(argname: str, value, expected_type, memo: Optional[_CallMemo]) -> None:
+def check_class(argname: str, value, expected_type, memo: _TypeCheckMemo) -> None:
     if not isclass(value):
         raise TypeError('type of {} must be a type; got {} instead'.format(
             argname, qualified_name(value)))
@@ -431,12 +434,9 @@ def check_class(argname: str, value, expected_type, memo: Optional[_CallMemo]) -
                 argname, qualified_name(expected_class), qualified_name(value)))
 
 
-def check_typevar(argname: str, value, typevar: TypeVar, memo: Optional[_CallMemo],
+def check_typevar(argname: str, value, typevar: TypeVar, memo: _TypeCheckMemo,
                   subclass_check: bool = False) -> None:
-    bound_type = None
-    if memo is not None:
-        bound_type = resolve_forwardref(memo.typevars.get(typevar, typevar.__bound__), memo)
-
+    bound_type = resolve_forwardref(memo.typevars.get(typevar, typevar.__bound__), memo)
     value_type = value if subclass_check else type(value)
     subject = argname if subclass_check else 'type of ' + argname
     if bound_type is None:
@@ -464,12 +464,12 @@ def check_typevar(argname: str, value, typevar: TypeVar, memo: Optional[_CallMem
                 '{} must be exactly {}; got {} instead'.
                 format(subject, qualified_name(bound_type), qualified_name(value_type)))
 
-    if memo is not None and typevar not in memo.typevars:
+    if typevar not in memo.typevars:
         # Bind the type variable to a concrete type
         memo.typevars[typevar] = value_type
 
 
-def check_literal(argname: str, value, expected_type, memo: Optional[_CallMemo]):
+def check_literal(argname: str, value, expected_type, memo: _TypeCheckMemo):
     try:
         args = expected_type.__args__
     except AttributeError:
@@ -505,9 +505,11 @@ def check_io(argname: str, value, expected_type):
 
 
 def check_protocol(argname: str, value, expected_type):
-    if not issubclass(type(value), expected_type):
-        raise TypeError('type of {} ({}) is not compatible with the {} protocol'.
-                        format(argname, type(value).__qualname__, expected_type.__qualname__))
+    # TODO: implement proper compatibility checking and support non-runtime protocols
+    if getattr(expected_type, '_is_runtime_protocol', False):
+        if not isinstance(value, expected_type):
+            raise TypeError('type of {} ({}) is not compatible with the {} protocol'.
+                            format(argname, type(value).__qualname__, expected_type.__qualname__))
 
 
 # Equality checks are applied to these
@@ -519,6 +521,7 @@ origin_type_checkers = {
     Dict: check_dict,
     list: check_list,
     List: check_list,
+    Literal: check_literal,
     Sequence: check_sequence,
     collections.abc.Sequence: check_sequence,
     collections.abc.Set: check_set,
@@ -527,15 +530,10 @@ origin_type_checkers = {
     tuple: check_tuple,
     Tuple: check_tuple,
     type: check_class,
+    Type: check_class,
     Union: check_union
 }
 _subclass_check_unions = hasattr(Union, '__union_set_params__')
-if Type is not None:
-    origin_type_checkers[Type] = check_class
-if Literal is not None:
-    origin_type_checkers[Literal] = check_literal
-if BPLiteral is not None:
-    origin_type_checkers[BPLiteral] = check_literal
 
 generator_origin_types = (Generator, collections.abc.Generator,
                           Iterator, collections.abc.Iterator,
@@ -548,7 +546,9 @@ if hasattr(collections.abc, 'AsyncGenerator'):
     asyncgen_origin_types += (collections.abc.AsyncGenerator,)
 
 
-def check_type(argname: str, value, expected_type, memo: Optional[_CallMemo] = None) -> None:
+def check_type(argname: str, value, expected_type, memo: Optional[_TypeCheckMemo] = None, *,
+               globals: Optional[Dict[str, Any]] = None,
+               locals: Optional[Dict[str, Any]] = None) -> None:
     """
     Ensure that ``value`` matches ``expected_type``.
 
@@ -559,6 +559,10 @@ def check_type(argname: str, value, expected_type, memo: Optional[_CallMemo] = N
     :param argname: name of the argument to check; used for error messages
     :param value: value to be checked against ``expected_type``
     :param expected_type: a class or generic type instance
+    :param globals: dictionary of global variables to use for resolving forward references
+        (defaults to the calling frame's globals)
+    :param locals: dictionary of local variables to use for resolving forward references
+        (defaults to the calling frame's locals)
 
     """
     if expected_type is Any or isinstance(value, Mock):
@@ -568,6 +572,16 @@ def check_type(argname: str, value, expected_type, memo: Optional[_CallMemo] = N
         # Only happens on < 3.6
         expected_type = type(None)
 
+    if memo is None:
+        frame = sys._getframe(1)
+        if globals is None:
+            globals = frame.f_globals
+        if locals is None:
+            locals = frame.f_locals
+
+        memo = _TypeCheckMemo(globals, locals)
+
+    expected_type = resolve_forwardref(expected_type, memo)
     origin_type = getattr(expected_type, '__origin__', None)
     if origin_type is not None:
         checker_func = origin_type_checkers.get(origin_type)
@@ -578,9 +592,6 @@ def check_type(argname: str, value, expected_type, memo: Optional[_CallMemo] = N
     elif isclass(expected_type):
         if issubclass(expected_type, Tuple):
             check_tuple(argname, value, expected_type, memo)
-        elif issubclass(expected_type, Callable) and hasattr(expected_type, '__args__'):
-            # Needed on Python 3.5.0 to 3.5.2
-            check_callable(argname, value, expected_type, memo)
         elif issubclass(expected_type, (float, complex)):
             check_number(argname, value, expected_type)
         elif _subclass_check_unions and issubclass(expected_type, Union):
@@ -608,7 +619,7 @@ def check_type(argname: str, value, expected_type, memo: Optional[_CallMemo] = N
     elif isinstance(expected_type, TypeVar):
         # Only happens on < 3.6
         check_typevar(argname, value, expected_type, memo)
-    elif BPLiteral is not None and isinstance(expected_type, BPLiteral.__class__):
+    elif isinstance(expected_type, Literal.__class__):
         # Only happens on < 3.7 when using Literal from typing_extensions
         check_literal(argname, value, expected_type, memo)
     elif (isfunction(expected_type) and
@@ -641,6 +652,9 @@ def check_return_type(retval, memo: Optional[_CallMemo] = None) -> bool:
         memo = _CallMemo(func, frame.f_locals)
 
     if 'return' in memo.type_hints:
+        if memo.type_hints['return'] is NoReturn:
+            raise TypeError('{}() was declared never to return but it did'.format(memo.func_name))
+
         try:
             check_type('the return value', retval, memo.type_hints['return'], memo)
         except TypeError as exc:  # suppress unnecessarily long tracebacks
